/*
 * Host AP (software wireless LAN access point) user space daemon for
 * Host AP kernel driver / Accounting
 * Copyright (c) 2002-2003, Jouni Malinen <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>


#include "hostapd.h"
#include "radius.h"
#include "eloop.h"
#include "accounting.h"
#include "driver.h"


static inline int
accounting_retransmit(hostapd *hapd, struct accounting_list *entry, time_t now)
{
	/* retransmit; remove entry if too many attempts */
	entry->attempts++;
	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "Resending RADIUS message (id=%d)"
		      " to accounting server\n", entry->msg->hdr->identifier);

	if (send(hapd->acct_serv_sock, entry->msg->buf, entry->msg->buf_used,
		 0) < 0) {
		perror("send");
		printf("Could not send accounting message to RADIUS server\n");
	}

	entry->next_try = now + entry->next_wait;
	entry->next_wait *= 2;
	if (entry->next_wait > RADIUS_ACCT_MAX_WAIT)
		entry->next_wait = RADIUS_ACCT_MAX_WAIT;
	if (entry->attempts >= RADIUS_ACCT_MAX_RETRIES) {
		printf("Removing un-ACKed Accounting-Request due to too many "
		       "failed retransmit attempts\n");
		return 1;
	}

	return 0;
}


static void accounting_list_timer(void *eloop_ctx, void *timeout_ctx)
{
	hostapd *hapd = eloop_ctx;
	time_t now, first;
	struct accounting_list *entry, *prev, *tmp;

	entry = hapd->acct_retransmit_list;
	if (!entry)
		return;

	time(&now);
	first = 0;

	prev = NULL;
	while (entry) {
		if (now >= entry->next_try &&
		    accounting_retransmit(hapd, entry, now)) {
			if (prev)
				prev->next = entry->next;
			else
				hapd->acct_retransmit_list = entry->next;

			tmp = entry;
			entry = entry->next;
			radius_msg_free(tmp->msg);
			free(tmp->msg);
			free(tmp);
			hapd->acct_retransmit_list_len--;
			continue;
		}

		if (first == 0 || entry->next_try < first)
			first = entry->next_try;

		prev = entry;
		entry = entry->next;
	}

	if (hapd->acct_retransmit_list) {
		if (first < now)
			first = now;
		eloop_register_timeout(first - now, 0,
				       accounting_list_timer, hapd, NULL);
	}
}


static void accounting_list_add(hostapd *hapd, struct radius_msg *msg)
{
	struct accounting_list *entry, *prev;

	if (eloop_terminated()) {
		/* No point in adding entries to retransmit queue since event
		 * loop has already been terminated. */
		radius_msg_free(msg);
		free(msg);
		return;
	}

	entry = malloc(sizeof(*entry));
	if (entry == NULL) {
		printf("Failed to add Accounting-Request into retransmit "
		       "list\n");
		radius_msg_free(msg);
		free(msg);
		return;
	}

	memset(entry, 0, sizeof(*entry));
	entry->msg = msg;
	time(&entry->first_try);
	entry->next_try = entry->first_try + RADIUS_ACCT_FIRST_WAIT;
	entry->attempts = 1;
	entry->next_wait = RADIUS_ACCT_FIRST_WAIT * 2;

	if (!hapd->acct_retransmit_list)
		eloop_register_timeout(RADIUS_ACCT_FIRST_WAIT, 0,
				       accounting_list_timer, hapd, NULL);

	entry->next = hapd->acct_retransmit_list;
	hapd->acct_retransmit_list = entry;

	if (hapd->acct_retransmit_list_len >= RADIUS_ACCT_MAX_ENTRIES) {
		printf("Removing the oldest un-ACKed Account-Request due to "
		       "retransmit list limits.\n");
		prev = NULL;
		while (entry->next) {
			prev = entry;
			entry = entry->next;
		}
		if (prev) {
			prev->next = NULL;
			radius_msg_free(entry->msg);
			free(entry->msg);
		}
	} else
		hapd->acct_retransmit_list_len++;
}


static struct accounting_list * accounting_list_get(hostapd *hapd, u8 id)
{
	struct accounting_list *entry;

	entry = hapd->acct_retransmit_list;
	while (entry) {
		if (entry->msg->hdr->identifier == id)
			break;
		entry = entry->next;
	}
	return entry;
}


static int accounting_list_del(hostapd *hapd, struct accounting_list *del)
{
	struct accounting_list *entry, *prev;

	prev = NULL;
	entry = hapd->acct_retransmit_list;
	while (entry) {
		if (entry == del) {
			if (prev)
				prev->next = entry->next;
			else
				hapd->acct_retransmit_list = entry->next;

			radius_msg_free(entry->msg);
			free(entry->msg);
			free(entry);
			hapd->acct_retransmit_list_len--;
			return 0;
		}
		prev = entry;
		entry = entry->next;
	}
	return -1;
}


static void accounting_list_free(hostapd *hapd)
{
	struct accounting_list *entry, *prev;

	entry = hapd->acct_retransmit_list;
	hapd->acct_retransmit_list = NULL;
	hapd->acct_retransmit_list_len = 0;
	while (entry) {
		radius_msg_free(entry->msg);
		free(entry->msg);
		prev = entry;
		entry = entry->next;
		free(prev);
	}
}


static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
					  int status_type)
{
	struct radius_msg *msg;
	char buf[128];
	u8 radius_id;
	u8 *val;
	size_t len;

	radius_id = hapd->next_radius_identifier++;
	msg = radius_msg_new(RADIUS_CODE_ACCOUNTING_REQUEST, radius_id);
	if (msg == NULL) {
		printf("Could not create net RADIUS packet\n");
		return NULL;
	}

	radius_msg_make_authenticator(msg, (u8 *) sta, sizeof(sta));

	snprintf(buf, sizeof(buf), "%08X-%08X",
		 hapd->acct_session_id_hi, sta->acct_session_id_lo);
	if (!radius_msg_add_attr(msg, RADIUS_ATTR_ACCT_SESSION_ID,
				 buf, strlen(buf))) {
		printf("Could not add Acct-Session-Id\n");
		goto fail;
	}

	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_ACCT_STATUS_TYPE,
				       status_type)) {
		printf("Could not add Acct-Status-Type\n");
		goto fail;
	}

	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_ACCT_AUTHENTIC,
				       hapd->conf->ieee802_1x ?
				       RADIUS_ACCT_AUTHENTIC_RADIUS :
				       RADIUS_ACCT_AUTHENTIC_LOCAL)) {
		printf("Could not add Acct-Authentic\n");
		goto fail;
	}

	val = sta->identity;
	len = sta->identity_len;
	if (!val) {
		snprintf(buf, sizeof(buf), RADIUS_ADDR_FORMAT,
			 MAC2STR(sta->addr));
		val = buf;
		len = strlen(val);
	}

	if (!radius_msg_add_attr(msg, RADIUS_ATTR_USER_NAME, val, len)) {
		printf("Could not add User-Name\n");
		goto fail;
	}

	if (!radius_msg_add_attr(msg, RADIUS_ATTR_NAS_IP_ADDRESS,
				 (u8 *) &hapd->conf->own_ip_addr, 4)) {
		printf("Could not add NAS-IP-Address\n");
		goto fail;
	}

	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_NAS_PORT, sta->aid)) {
		printf("Could not add NAS-Port\n");
		goto fail;
	}

	snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT ":%s",
		 MAC2STR(hapd->own_addr), hapd->conf->ssid);
	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLED_STATION_ID,
				 buf, strlen(buf))) {
		printf("Could not add Called-Station-Id\n");
		goto fail;
	}

	snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT,
		 MAC2STR(sta->addr));
	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLING_STATION_ID,
				 buf, strlen(buf))) {
		printf("Could not add Calling-Station-Id\n");
		goto fail;
	}

	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_NAS_PORT_TYPE,
				       RADIUS_NAS_PORT_TYPE_IEEE_802_11)) {
		printf("Could not add NAS-Port-Type\n");
		goto fail;
	}

	snprintf(buf, sizeof(buf), "CONNECT 11Mbps 802.11b");
	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CONNECT_INFO,
				 buf, strlen(buf))) {
		printf("Could not add Connect-Info\n");
		goto fail;
	}

	return msg;

 fail:
	radius_msg_free(msg);
	free(msg);
	return NULL;
}


static void accounting_send(hostapd *hapd, struct radius_msg *msg)
{
	radius_msg_finish_acct(msg, hapd->conf->acct_server_shared_secret,
			       hapd->conf->acct_server_shared_secret_len);

	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
		      "Sending RADIUS message to accounting server\n");
	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MSGDUMPS))
		radius_msg_dump(msg);

	if (send(hapd->acct_serv_sock, msg->buf, msg->buf_used, 0) < 0) {
		perror("send");
		printf("Could not send accounting message to RADIUS server\n");
	}

	accounting_list_add(hapd, msg);
}


void accounting_sta_start(hostapd *hapd, struct sta_info *sta)
{
	struct radius_msg *msg;

	if (!hapd->conf->acct_server_addr.s_addr)
		return;

	time(&sta->acct_session_start);

	msg = accounting_msg(hapd, sta, RADIUS_ACCT_STATUS_TYPE_START);
	if (msg)
		accounting_send(hapd, msg);
}


void accounting_sta_stop(hostapd *hapd, struct sta_info *sta)
{
	struct radius_msg *msg;
	int cause = sta->acct_terminate_cause;
	struct hostap_sta_driver_data data;

	if (!hapd->conf->acct_server_addr.s_addr)
		return;

	msg = accounting_msg(hapd, sta, RADIUS_ACCT_STATUS_TYPE_STOP);
	if (!msg) {
		printf("Could not create RADIUS Accounting message\n");
		return;
	}

	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_ACCT_SESSION_TIME,
				       time(NULL) - sta->acct_session_start)) {
		printf("Could not add Acct-Session-Time\n");
		goto fail;
	}

	if (hostapd_read_sta_driver_data(hapd, &data, sta->addr) == 0) {
		if (!radius_msg_add_attr_int32(msg,
					       RADIUS_ATTR_ACCT_INPUT_PACKETS,
					       data.rx_packets)) {
			printf("Could not add Acct-Input-Packets\n");
			goto fail;
		}
		if (!radius_msg_add_attr_int32(msg,
					       RADIUS_ATTR_ACCT_OUTPUT_PACKETS,
					       data.tx_packets)) {
			printf("Could not add Acct-Output-Packets\n");
			goto fail;
		}
		if (!radius_msg_add_attr_int32(msg,
					       RADIUS_ATTR_ACCT_INPUT_OCTETS,
					       data.rx_bytes)) {
			printf("Could not add Acct-Input-Octets\n");
			goto fail;
		}
		if (!radius_msg_add_attr_int32(msg,
					       RADIUS_ATTR_ACCT_OUTPUT_OCTETS,
					       data.tx_bytes)) {
			printf("Could not add Acct-Output-Octets\n");
			goto fail;
		}
	}

	if (eloop_terminated())
		cause = RADIUS_ACCT_TERMINATE_CAUSE_ADMIN_REBOOT;

	if (cause &&
	    !radius_msg_add_attr_int32(msg, RADIUS_ATTR_ACCT_TERMINATE_CAUSE,
				       cause)) {
		printf("Could not add Acct-Terminate-Cause\n");
		goto fail;
	}


	if (msg)
		accounting_send(hapd, msg);
	return;

 fail:
	radius_msg_free(msg);
	free(msg);
}


/* Process the RADIUS frames from Accounting Server */
static void accounting_receive(int sock, void *eloop_ctx, void *sock_ctx)
{
	hostapd *hapd = (hostapd *) eloop_ctx;
	int len;
	unsigned char buf[3000];
	struct radius_msg *msg;
	struct accounting_list *req;

	len = recv(sock, buf, sizeof(buf), 0);
	if (len < 0) {
		perror("recv");
		return;
	}
	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
		      "Received %d bytes from accounting server\n", len);
	if (len == sizeof(buf)) {
		printf("Possibly too long UDP frame for our buffer - "
		       "dropping it\n");
		return;
	}

	msg = radius_msg_parse(buf, len);
	if (msg == NULL) {
		printf("Parsing incoming RADIUS frame failed\n");
		return;
	}

	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
		      "Received RADIUS message\n");
	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MSGDUMPS))
		radius_msg_dump(msg);

	if (msg->hdr->code != RADIUS_CODE_ACCOUNTING_RESPONSE) {
		printf("Unknown RADIUS message code\n");
		goto done;
	}

	req = accounting_list_get(hapd, msg->hdr->identifier);

	if (!req) {
		printf("No matching Accounting-Request found for id %d\n",
		       msg->hdr->identifier);
		goto done;
	}

	if (radius_msg_verify_acct(msg, hapd->conf->acct_server_shared_secret,
				   hapd->conf->acct_server_shared_secret_len,
				   req->msg)) {
		printf("Incoming RADIUS packet did not have correct "
		       "Authenticator - dropped\n");
		goto done;
	}

	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
		      "Removing ACKed Accounting-Request (id=%d) from "
		      "retransmit list\n", req->msg->hdr->identifier);
	accounting_list_del(hapd, req);

 done:
	if (msg) {
		radius_msg_free(msg);
		free(msg);
	}
}


int accounting_init(hostapd *hapd)
{
	struct sockaddr_in serv;

	/* Acct-Session-Id should be unique over reboots. If reliable clock is
	 * not available, this could be replaced with reboot counter, etc. */
	hapd->acct_session_id_hi = time(NULL);

	if (!hapd->conf->acct_server_addr.s_addr)
		return 0;

	hapd->acct_serv_sock = socket(PF_INET, SOCK_DGRAM, 0);
	if (hapd->acct_serv_sock < 0) {
		perror("socket[PF_INET,SOCK_DGRAM]");
		return -1;
	}

	memset(&serv, 0, sizeof(serv));
	serv.sin_family = AF_INET;
	serv.sin_addr.s_addr = hapd->conf->acct_server_addr.s_addr;
	serv.sin_port = htons(hapd->conf->acct_server_port);

	if (connect(hapd->acct_serv_sock, (struct sockaddr *) &serv,
		    sizeof(serv)) < 0) {
		perror("connect[acct_serv_sock]");
		return -1;
	}

	if (eloop_register_read_sock(hapd->acct_serv_sock,
				     accounting_receive, hapd, NULL)) {
		printf("Could not register read socket for accounting "
		       "server\n");
		return -1;
	}

	return 0;
}


void accounting_deinit(hostapd *hapd)
{
	if (hapd->acct_serv_sock >= 0)
		close(hapd->acct_serv_sock);
	accounting_list_free(hapd);
}
